@@ -15,46 +15,36 @@
 #with this program.  If not, see <http://www.gnu.org/licenses/>.
 ### END LICENSE
 
-import sys
+import argparse
+import cgi
+import contextlib
+import html
+import json
+import logging
+import math
+import os
 import re
-import os, time
-import logging, argparse
 import signal
+import sys
+import time
+import urllib.error
+import urllib.parse
+import urllib.request
 
 import gi
 gi.require_version('Gst', '1.0')
 from gi.repository import Gst, GstPbutils, GObject, Gtk, Gdk, Pango, GdkPixbuf, Gio, GLib
-import contextlib
-import html
-import math
-import urllib.request, urllib.error, urllib.parse
-import json
-if sys.platform != 'win32':
-    from dbus.mainloop.glib import DBusGMainLoop
-    DBusGMainLoop(set_as_default=True)
 
-# Check if we are working in the source tree or from the installed
-# package and mangle the python path accordingly
-realPath = os.path.realpath(sys.argv[0])  # If this file is run from a symlink, it needs to follow the symlink
-if os.path.dirname(realPath) != ".":
-    if realPath[0] == "/":
-        fullPath = os.path.dirname(realPath)
-    else:
-        fullPath = os.getcwd() + "/" + os.path.dirname(realPath)
-else:
-    fullPath = os.getcwd()
-sys.path.insert(0, os.path.dirname(fullPath))
+if Gtk.get_major_version() < 3 or Gtk.get_minor_version() < 14:
+    sys.exit('Gtk 3.14 is required')
 
 from . import AboutPithosDialog, PreferencesPithosDialog, StationsDialog
-from .util import parse_proxy, open_browser
-from .pithosconfig import get_ui_file, get_media_file, VERSION
 from .gobject_worker import GObjectWorker
-from .plugin import load_plugins
-if sys.platform != 'win32':
-    from .dbus_service import PithosDBusProxy
-    from .mpris import PithosMprisService
 from .pandora import *
 from .pandora.data import *
+from .pithosconfig import get_ui_file, get_media_file, VERSION
+from .plugin import load_plugins
+from .util import parse_proxy, open_browser
 
 pacparser_imported = False
 try:
@@ -63,16 +53,6 @@ try:
 except ImportError:
     pass
 
-def buttonMenu(button, menu):
-    def cb(button):
-        allocation = button.get_allocation()
-        x, y = button.get_window().get_origin()[1:]
-        x += allocation.x
-        y += allocation.y + allocation.height
-        menu.popup(None, None, (lambda *ignore: (x, y, True)), None, 1, Gtk.get_current_event_time())
-
-    button.connect('clicked', cb)
-
 ALBUM_ART_SIZE = 96
 ALBUM_ART_X_PAD = 6
 
@@ -104,8 +84,7 @@ class CellRendererAlbumArt(Gtk.CellRenderer):
             Gdk.cairo_set_source_pixbuf(ctx, self.rate_bg, x, y)
             ctx.paint()
 
-            icon = widget.get_style_context().lookup_icon_set(self.icon)
-            pixbuf = icon.render_icon_pixbuf(widget.get_style_context(), Gtk.IconSize.MENU)
+            pixbuf = Gtk.IconTheme.get_default().load_icon(self.icon, Gtk.IconSize.MENU, 0)
             x = cell_area.x+(cell_area.width-pixbuf.get_width())-5 - ALBUM_ART_X_PAD # right
             y = cell_area.y+(cell_area.height-pixbuf.get_height())-5 # bottom
             Gdk.cairo_set_source_pixbuf(ctx, pixbuf, x, y)
@@ -124,6 +103,16 @@ def get_album_art(url, *extra):
         loader.write(image)
         return (loader.get_pixbuf(),) + extra
 
+class PlayerStatus (object):
+  def __init__(self):
+    self.reset()
+
+  def reset(self):
+    self.async_done = False
+    self.began_buffering = None
+    self.buffer_percent = 0
+    self.pending_duration_query = False
+
 
 class PithosWindow(Gtk.ApplicationWindow):
     __gtype_name__ = "PithosWindow"
@@ -174,10 +163,7 @@ class PithosWindow(Gtk.ApplicationWindow):
 
         self.plugins = {}
         load_plugins(self)
-
-        if sys.platform != 'win32':
-            self.dbus_service = PithosDBusProxy(self)
-            self.mpris = PithosMprisService(self)
+        self.prefs_dlg.set_plugins(self.plugins)
 
         if not self.preferences['username']:
             self.show_preferences(is_startup=True)
@@ -194,10 +180,14 @@ class PithosWindow(Gtk.ApplicationWindow):
         self.stations_model = Gtk.ListStore(GObject.TYPE_PYOBJECT, str)
 
         Gst.init(None)
+        self._query_duration = Gst.Query.new_duration(Gst.Format.TIME)
+        self._query_position = Gst.Query.new_position(Gst.Format.TIME)
         self.player = Gst.ElementFactory.make("playbin", "player");
-        self.player.props.flags |= (1 << 7) # enable progressive download (GST_PLAY_FLAG_DOWNLOAD)
+
         bus = self.player.get_bus()
         bus.add_signal_watch()
+        bus.connect("message::async-done", self.on_gst_async_done)
+        bus.connect("message::duration-changed", self.on_gst_duration_changed)
         bus.connect("message::eos", self.on_gst_eos)
         bus.connect("message::buffering", self.on_gst_buffering)
         bus.connect("message::error", self.on_gst_error)
@@ -205,16 +195,16 @@ class PithosWindow(Gtk.ApplicationWindow):
         bus.connect("message::tag", self.on_gst_tag)
         self.player.connect("notify::volume", self.on_gst_volume)
         self.player.connect("notify::source", self.on_gst_source)
-        self.time_format = Gst.Format.TIME
+
+        self.player_status = PlayerStatus()
 
         self.stations_dlg = None
 
-        self.playing = False
+        self.playing = None # None is a special "Waiting to play" state
         self.current_song_index = None
         self.current_station = None
         self.current_station_id = self.preferences.get('last_station_id')
 
-        self.buffer_percent = 100
         self.auto_retrying_auth = False
         self.have_stations = False
         self.playcount = 0
@@ -223,7 +213,7 @@ class PithosWindow(Gtk.ApplicationWindow):
         self.gstreamer_error = ''
         self.waiting_for_playlist = False
         self.start_new_playlist = False
-
+        self.ui_loop_timer_id = 0
         self.worker = GObjectWorker()
         self.art_worker = GObjectWorker()
 
@@ -462,14 +452,14 @@ class PithosWindow(Gtk.ApplicationWindow):
             return self.next_song()
 
         logging.info("Starting song: index = %i"%(song_index))
-        self.buffer_percent = 0
-        self.song_started = False
+        self.player_status.reset()
+
         self.player.set_property("uri", self.current_song.audioUrl)
-        self.play()
+        self.player.set_state(Gst.State.PAUSED)
+        self.playing = None
         self.playcount += 1
 
         self.current_song.start_time = time.time()
-
         self.songs_treeview.scroll_to_cell(song_index, use_align=True, row_align = 1.0)
         self.songs_treeview.set_cursor(song_index, None, 0)
         self.set_title("Pithos - %s by %s" % (self.current_song.title, self.current_song.artist))
@@ -481,14 +471,13 @@ class PithosWindow(Gtk.ApplicationWindow):
 
     def user_play(self, *ignore):
         self.play()
-        self.song_started = True
         self.emit('user-changed-play-state', True)
 
     def play(self):
         if not self.playing:
             self.playing = True
+            self.create_ui_loop()
         self.player.set_state(Gst.State.PLAYING)
-        GLib.timeout_add_seconds(1, self.update_song_row)
         self.playpause_image.set_from_icon_name('media-playback-pause-symbolic', Gtk.IconSize.SMALL_TOOLBAR)
         self.update_song_row()
         self.emit('play-state-changed', True)
@@ -499,6 +488,7 @@ class PithosWindow(Gtk.ApplicationWindow):
 
     def pause(self):
         self.playing = False
+        self.destroy_ui_loop()
         self.player.set_state(Gst.State.PAUSED)
         self.playpause_image.set_from_icon_name('media-playback-start-symbolic', Gtk.IconSize.SMALL_TOOLBAR)
         self.update_song_row()
@@ -509,19 +499,17 @@ class PithosWindow(Gtk.ApplicationWindow):
         prev = self.current_song
         if prev and prev.start_time:
             prev.finished = True
-            dur_stat, dur = self.player.query_duration(self.time_format)
-            prev.duration = dur//1000000000 if dur_stat else None
-            pos_stat, pos = self.player.query_position(self.time_format)
-            prev.position = pos//1000000000 if pos_stat else None
+            prev.position = self.query_position()
             self.emit("song-ended", prev)
 
         self.playing = False
+        self.destroy_ui_loop()
         self.player.set_state(Gst.State.NULL)
         self.emit('play-state-changed', False)
 
     def user_playpause(self, *ignore):
         self.playpause_notify()
-        
+
     def playpause(self, *ignore):
         logging.info("playpause")
         if self.playing:
@@ -604,7 +592,7 @@ class PithosWindow(Gtk.ApplicationWindow):
         dialog.props.secondary_text = submsg
         dialog.set_default_response(1)
 
-        response = dialog.run()
+        dialog.run()
         dialog.hide()
 
         self.quit()
@@ -634,6 +622,35 @@ class PithosWindow(Gtk.ApplicationWindow):
             self.get_playlist(start = True)
         self.stations_combo.set_active(self.station_index(station))
 
+    def query_position(self):
+      pos_stat = self.player.query(self._query_position)
+      if pos_stat:
+        _, position = self._query_position.parse_position()
+        return position
+
+    def query_duration(self):
+      dur_stat = self.player.query(self._query_duration)
+      if dur_stat:
+        _, duration = self._query_duration.parse_duration()
+        return duration
+
+    def on_gst_async_done(self, bus, message):
+      logging.debug("on_gst_async_done")
+      self.player_status.async_done = True
+      if self.player_status.pending_duration_query:
+        self.current_song.duration = self.query_duration()
+        self.current_song.duration_message = self.format_time(self.current_song.duration)
+        self.check_if_song_is_ad()
+        self.player_status.pending_duration_query = False
+
+    def on_gst_duration_changed(self, bus, message):
+      if self.player_status.async_done:
+        self.current_song.duration = self.query_duration()
+        self.current_song.duration_message = self.format_time(self.current_song.duration)
+        self.check_if_song_is_ad()
+      else:
+        self.player_status.pending_duration_query = True
+
     def on_gst_eos(self, bus, message):
         logging.info("EOS")
         self.next_song()
@@ -669,7 +686,7 @@ class PithosWindow(Gtk.ApplicationWindow):
 
     def gst_tag_handler(self, tag_info):
         def handler(_x, tag, _y):
-            # An exhaustive list of tags is available at 
+            # An exhaustive list of tags is available at
             # https://developer.gnome.org/gstreamer/stable/gstreamer-GstTagList.html
             # but Pandora seems to only use those
             if tag == 'datetime':
@@ -684,32 +701,24 @@ class PithosWindow(Gtk.ApplicationWindow):
 
             logging.debug('Found tag "%s" in stream: "%s" (type: %s)' % (tag, value, type(value)))
 
-            if tag == 'audio-codec':
-                # At that point we should have duration information, check for ads
-                self.check_if_song_is_ad()
-
             if tag == 'bitrate':
-                self.current_song.bitrate = value
+                self.current_song.bitrate = value / 1000
                 self.update_song_row()
 
         return handler
 
     def check_if_song_is_ad(self):
         if self.current_song.is_ad is None:
-            dur_stat, dur_int = self.player.query_duration(self.time_format)
-
-            if not dur_stat:
-                logging.warning('dur_stat is False. The assumption that duration is available once the audio-codec messages feeds is bad.')
-            else:
-                dur_int /= 1e9
-
-                if dur_int < 45.0:  # Less than 45 seconds we assume it's an ad
+            if self.current_song.duration:
+                if self.current_song.get_duration_sec() < 45:  # Less than 45 seconds we assume it's an ad
                     logging.info('Ad detected!')
                     self.current_song.is_ad = True
                     self.update_song_row()
                 else:
                     logging.info('Not an Ad..')
                     self.current_song.is_ad = False
+            else:
+                logging.warning('dur_stat is False. The assumption that duration is available once the audio-codec messages feeds is bad.')
 
     def on_gst_tag(self, bus, message):
         tag_info = message.parse_tag()
@@ -721,19 +730,31 @@ class PithosWindow(Gtk.ApplicationWindow):
         # Note that applications should keep/set the pipeline in the PAUSED state when a BUFFERING
         # message is received with a buffer percent value < 100 and set the pipeline back to PLAYING
         # state when a BUFFERING message with a value of 100 percent is received.
-        
+
         # 100% doesn't mean the entire song is downloaded, but it does mean that it's safe to play.
         # trying to play before 100% will cause stuttering.
         percent = message.parse_buffering()
-        self.buffer_percent = percent
+        logging.debug("Buffering (%i%%)", percent)
+
         if percent < 100:
-            self.player.set_state(Gst.State.PAUSED)
+            # If our previous buffer was at 100, but now it's < 100,
+            # then we should pause until the buffer is full.
+            if self.player_status.buffer_percent == 100:
+                logging.debug("Buffer underrun. Pausing pipeline")
+                self.player.set_state(Gst.State.PAUSED)
+                self.player_status.began_buffering = time.time()
         else:
-            if self.playing:
+            if self.playing is None: # Not playing but waiting to
+                logging.debug("Buffer 100%. Song starting")
                 self.play()
-                self.song_started = True
+            elif self.playing:
+                logging.debug("Buffer recovery. Restarting pipeline")
+                self.player.set_state(Gst.State.PLAYING)
+            else:
+                logging.debug("Buffer recovery. User paused")
+            self.player_status.began_buffering = None
+        self.player_status.buffer_percent = percent
         self.update_song_row()
-        logging.debug("Buffering (%i%%)"%self.buffer_percent)
 
     def set_volume_cb(self, volume):
         # Convert to the cubic scale that the volume slider uses
@@ -763,20 +784,17 @@ class PithosWindow(Gtk.ApplicationWindow):
         album = html.escape(song.album)
         msg = []
         if song is self.current_song:
-            dur_stat, dur_int = self.player.query_duration(self.time_format)
-            pos_stat, pos_int = self.player.query_position(self.time_format)
-            if not self.song_started:
-                pos_int = 0
+            song.position = self.query_position()
             if not song.bitrate is None:
-                msg.append("%0dkbit/s" % (song.bitrate / 1000))
-            if dur_stat and pos_stat:
-                dur_str = self.format_time(dur_int)
-                pos_str = self.format_time(pos_int)
-                msg.append("%s / %s" %(pos_str, dur_str))
+                msg.append("%0dkbit/s" % (song.bitrate))
+
+            if song.position is not None and song.duration is not None:
+                pos_str = self.format_time(song.position)
+                msg.append("%s / %s" % (pos_str, song.duration_message))
                 if not self.playing:
                     msg.append("Paused")
-            if self.buffer_percent < 100:
-                msg.append("Buffering (%i%%)"%self.buffer_percent)
+            if self.player_status.buffer_percent < 100:
+                msg.append("Buffering (%i%%)" % self.player_status.buffer_percent)
         if song.message:
             msg.append(song.message)
         msg = " - ".join(msg)
@@ -792,11 +810,11 @@ class PithosWindow(Gtk.ApplicationWindow):
 
     def song_icon(self, song):
         if song.tired:
-            return Gtk.STOCK_JUMP_TO
+            return 'go-jump'
         if song.rating == RATE_LOVE:
-            return Gtk.STOCK_ABOUT
+            return 'emblem-favorite'
         if song.rating == RATE_BAN:
-            return Gtk.STOCK_CANCEL
+            return 'dialog-error'
 
     def update_song_row(self, song = None):
         if song is None:
@@ -806,12 +824,24 @@ class PithosWindow(Gtk.ApplicationWindow):
             self.songs_model[song.index][2] = self.song_icon(song) or ""
         return self.playing
 
+    def create_ui_loop(self):
+        if not self.ui_loop_timer_id:
+            self.ui_loop_timer_id = GLib.timeout_add_seconds(1, self.update_song_row)
+
+    def destroy_ui_loop(self):
+        if self.ui_loop_timer_id:
+            GLib.source_remove(self.ui_loop_timer_id)
+            self.ui_loop_timer_id = 0
+
     def stations_combo_changed(self, widget):
         index = widget.get_active()
         if index>=0:
             self.station_changed(self.stations_model[index][0])
 
     def format_time(self, time_int):
+        if time_int is None:
+          return None
+
         time_int = time_int // 1000000000
         s = time_int % 60
         time_int //= 60
@@ -919,7 +949,6 @@ class PithosWindow(Gtk.ApplicationWindow):
                 self.start_song(self.selected_song().index)
 
     def set_player_volume(self, value):
-        logging.info('%.3f' % value)
         # Use a cubic scale for volume. This matches what PulseAudio uses.
         volume = math.pow(value, 3)
         self.player.set_property("volume", volume)
@@ -943,7 +972,7 @@ class PithosWindow(Gtk.ApplicationWindow):
         about = AboutPithosDialog.NewAboutPithosDialog()
         about.set_transient_for(self)
         about.set_version(VERSION)
-        response = about.run()
+        about.run()
         about.destroy()
 
     def show_preferences(self, is_startup=False):
@@ -969,7 +998,6 @@ class PithosWindow(Gtk.ApplicationWindow):
                         self.pandora_connect()
             else:
                 self.prefs_dlg.set_type_hint(Gdk.WindowTypeHint.DIALOG)
-            load_plugins(self)
 
     def show_stations(self):
         if self.stations_dlg:
